# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.gdp import GDP_Error, Disjunction
from pyomo.gdp.disjunct import DisjunctData, Disjunct

import pyomo.core.expr as EXPR
from pyomo.core import (
    Block,
    Suffix,
    TraversalStrategy,
    SortComponents,
    LogicalConstraint,
    value,
)
from pyomo.core.base.block import BlockData
from pyomo.common.collections import ComponentMap, ComponentSet, OrderedSet
from pyomo.opt import TerminationCondition, SolverStatus

from weakref import ref as weakref_ref
from collections import defaultdict
import logging

logger = logging.getLogger('pyomo.gdp')

_acceptable_termination_conditions = set(
    [
        TerminationCondition.optimal,
        TerminationCondition.globallyOptimal,
        TerminationCondition.locallyOptimal,
    ]
)
_infeasible_termination_conditions = set(
    [TerminationCondition.infeasible, TerminationCondition.invalidProblem]
)


class NORMAL:
    pass


class INFEASIBLE:
    pass


class NONOPTIMAL:
    pass


def verify_successful_solve(results):
    status = results.solver.status
    term = results.solver.termination_condition

    if status == SolverStatus.ok and term in _acceptable_termination_conditions:
        return NORMAL
    elif term in _infeasible_termination_conditions:
        return INFEASIBLE
    else:
        return NONOPTIMAL


def clone_without_expression_components(expr, substitute=None):
    """A function that is used to clone an expression.

    Cloning is roughly equivalent to calling ``copy.deepcopy``.
    However, the :attr:`clone_leaves` argument can be used to
    clone only interior (i.e. non-leaf) nodes in the expression
    tree.   Note that named expression objects are treated as
    leaves when :attr:`clone_leaves` is :const:`True`, and hence
    those subexpressions are not cloned.

    This function uses a non-recursive
    logic, which makes it more scalable than the logic in
    ``copy.deepcopy``.

    Args:
        expr: The expression that will be cloned.
        substitute (dict): A dictionary mapping object ids to
            objects.  This dictionary has the same semantics as
            the memo object used with ``copy.deepcopy``.  Defaults
            to None, which indicates that no user-defined
            dictionary is used.

    Returns:
        The cloned expression.
    """
    if substitute is None:
        substitute = {}
    #
    visitor = EXPR.ExpressionReplacementVisitor(
        substitute=substitute, remove_named_expressions=True
    )
    return visitor.walk_expression(expr)


def _raise_disjunct_in_multiple_disjunctions_error(disjunct, disjunction):
    # we've transformed it, which means this is the second time it's appearing
    # in a Disjunction
    raise GDP_Error(
        "The disjunct '%s' has been transformed, but '%s', a disjunction "
        "it appears in, has not. Putting the same disjunct in "
        "multiple disjunctions is not supported." % (disjunct.name, disjunction.name)
    )


class GDPTree:
    """
    Stores a forest representing the hierarchy between GDP components on a
    model: for single-level GDPs, each tree is rooted at a Disjunction and
    each of the Disjuncts in the Disjunction is a leaf. For nested GDPs, the
    Disjuncts may not be leaves, and could have child Disjunctions of their
    own.
    """

    def __init__(self):
        self._children = {}
        self._in_degrees = {}
        # Every node has exactly one or 0 parents.
        self._parent = {}

        self._root_disjunct = {}
        # This needs to be ordered so that topological sort is deterministic
        self._vertices = OrderedSet()

    @property
    def vertices(self):
        return self._vertices

    def add_node(self, u):
        self._vertices.add(u)

    def parent(self, u):
        """Returns the parent node of u, or None if u is a root.

        Arg:
            u : A node in the tree
        """
        if u in self._parent:
            return self._parent[u]
        if u not in self._vertices:
            raise ValueError(
                "'%s' is not a vertex in the GDP tree. Cannot "
                "retrieve its parent." % u
            )
        else:
            return None

    def children(self, u):
        """Returns the direct descendents of node u.

        Arg:
            u : A node in the tree
        """
        return self._children[u]

    def parent_disjunct(self, u):
        """Returns the parent Disjunct of u, or None if u is the
        closest-to-root Disjunct in the forest.

        Arg:
            u : A node in the forest
        """
        if u.ctype is Disjunct:
            return self.parent(self.parent(u))
        else:
            return self.parent(u)

    def root_disjunct(self, u):
        """Returns the highest parent Disjunct in the hierarchy, or None if
        the component is not nested.

        Arg:
            u : A node in the tree
        """
        rootmost_disjunct = None
        parent = self.parent(u)
        while True:
            if parent is None:
                return rootmost_disjunct
            if parent.ctype is Disjunct:
                rootmost_disjunct = parent
            parent = self.parent(parent)

    def add_node(self, u):
        if u not in self._children:
            self._children[u] = OrderedSet()
        self._vertices.add(u)

    def add_edge(self, u, v):
        self.add_node(u)
        self.add_node(v)
        self._children[u].add(v)
        if v in self._parent and self._parent[v] is not u:
            _raise_disjunct_in_multiple_disjunctions_error(v, u)
        self._parent[v] = u

    def _visit_vertex(self, u, leaf_to_root):
        if u in self._children:
            for v in self._children[u]:
                if v not in leaf_to_root:
                    self._visit_vertex(v, leaf_to_root)
        # we're done--we've been to all its children
        leaf_to_root.add(u)

    def _reverse_topological_iterator(self):
        # this returns nodes of the tree ordered so that no node is before any
        # of its descendents.
        leaf_to_root = OrderedSet()
        for u in self.vertices:
            if u not in leaf_to_root:
                self._visit_vertex(u, leaf_to_root)

        return leaf_to_root

    def topological_sort(self):
        return list(reversed(self._reverse_topological_iterator()))

    def reverse_topological_sort(self):
        return self._reverse_topological_iterator()

    def in_degree(self, u):
        if u not in self._parent:
            return 0
        return 1

    def is_leaf(self, u):
        if len(self._children[u]) == 0:
            return True
        return False

    @property
    def leaves(self):
        for u, children in self._children.items():
            if len(children) == 0:
                yield u

    @property
    def disjunct_nodes(self):
        for v in self._vertices:
            if v.ctype is Disjunct:
                yield v


def _parent_disjunct(obj):
    parent = obj.parent_block()
    while parent is not None:
        if parent.ctype is Disjunct:
            return parent
        parent = parent.parent_block()

    return None


def _check_properly_deactivated(disjunct):
    if disjunct.indicator_var.is_fixed():
        if not value(disjunct.indicator_var):
            # The user cleanly deactivated the disjunct: there
            # is nothing for us to do here.
            return
        else:
            raise GDP_Error(
                "The disjunct '%s' is deactivated, but the "
                "indicator_var is fixed to %s. This makes no sense."
                % (disjunct.name, value(disjunct.indicator_var))
            )
    if disjunct._transformation_block is None:
        raise GDP_Error(
            "The disjunct '%s' is deactivated, but the "
            "indicator_var is not fixed and the disjunct does not "
            "appear to have been transformed. This makes no sense. "
            "(If the intent is to deactivate the disjunct, fix its "
            "indicator_var to False.)" % (disjunct.name,)
        )


def _gather_disjunctions(block, gdp_tree, include_root=True):
    if not include_root:
        # The argument 'block' may be a root node (it was in the list of
        # targets): We will not add it to the tree in this call. It may be added
        # later if in fact it is a descendent of another target, but as far as
        # we know now, it does not belong in the tree.
        root = block
    to_explore = [block]
    while to_explore:
        block = to_explore.pop()
        for disjunction in block.component_data_objects(
            Disjunction,
            active=True,
            sort=SortComponents.deterministic,
            descend_into=Block,
        ):
            # add the node (because it might be an empty Disjunction and block
            # might be a Block, in case it wouldn't get added below.)
            gdp_tree.add_node(disjunction)
            for disjunct in disjunction.disjuncts:
                if not disjunct.active:
                    if disjunct.transformation_block is not None:
                        _raise_disjunct_in_multiple_disjunctions_error(
                            disjunct, disjunction
                        )
                    _check_properly_deactivated(disjunct)
                    continue
                gdp_tree.add_edge(disjunction, disjunct)
                to_explore.append(disjunct)
            if block.ctype is Disjunct:
                if not include_root and block is root:
                    continue
                gdp_tree.add_edge(block, disjunction)

    return gdp_tree


def get_gdp_tree(targets, instance, knownBlocks=None):
    if knownBlocks is None:
        knownBlocks = {}
    gdp_tree = GDPTree()
    for t in targets:
        # first check it's not insane, that is, it is at least on the instance
        if not is_child_of(parent=instance, child=t, knownBlocks=knownBlocks):
            raise GDP_Error(
                "Target '%s' is not a component on instance "
                "'%s'!" % (t.name, instance.name)
            )
        if t.ctype is Block or isinstance(t, BlockData):
            _blocks = t.values() if t.is_indexed() else (t,)
            for block in _blocks:
                if not block.active:
                    continue
                gdp_tree = _gather_disjunctions(block, gdp_tree, include_root=False)
        elif t.ctype is Disjunction:
            parent = _parent_disjunct(t)
            if parent is not None and parent in targets:
                gdp_tree.add_edge(parent, t)
            _disjunctions = t.values() if t.is_indexed() else (t,)
            for disjunction in _disjunctions:
                if disjunction.algebraic_constraint is not None:
                    # It's already transformed.
                    continue
                gdp_tree.add_node(disjunction)
                for disjunct in disjunction.disjuncts:
                    if not disjunct.active:
                        if disjunct.transformation_block is not None:
                            _raise_disjunct_in_multiple_disjunctions_error(
                                disjunct, disjunction
                            )
                        _check_properly_deactivated(disjunct)
                        continue
                    gdp_tree.add_edge(disjunction, disjunct)
                    gdp_tree = _gather_disjunctions(disjunct, gdp_tree)
        else:
            # There's nothing else we care about, so we don't know how to
            # deal with this
            raise GDP_Error(
                "Target '%s' was not a Block, Disjunct, or Disjunction. "
                "It was of type %s and can't be transformed." % (t.name, type(t))
            )
    return gdp_tree


def preprocess_targets(targets, instance, knownBlocks, gdp_tree=None):
    if gdp_tree is None:
        gdp_tree = get_gdp_tree(targets, instance, knownBlocks)
    # this is for bigm: We need to transform from the leaves up, so we want a
    # reverse of a topological sort: no parent can come before its child.
    return gdp_tree.reverse_topological_sort()


# [ESJ 07/09/2019 Should this be a more general utility function elsewhere?  I'm
# putting it here for now so that all the gdp transformations can use it.
# Returns True if child is a node or leaf in the tree rooted at parent, False
# otherwise. Accepts list of known components in the tree and updates this list
# to enhance performance in future calls. Note that both child and parent must
# be blocks!
def is_child_of(parent, child, knownBlocks=None):
    # Note: we can get away with a dictionary and not ComponentMap because we
    # will only store Blocks (or their ilk), and Blocks are hashable (only
    # derivatives of NumericValue are not hashable)
    if knownBlocks is None:
        knownBlocks = {}
    tmp = set()
    node = child if isinstance(child, (Block, BlockData)) else child.parent_block()
    while True:
        known = knownBlocks.get(node)
        if known:
            knownBlocks.update({c: True for c in tmp})
            return True
        if known is not None and not known:
            knownBlocks.update({c: False for c in tmp})
            return False
        if node is parent:
            knownBlocks.update({c: True for c in tmp})
            return True
        if node is None:
            knownBlocks.update({c: False for c in tmp})
            return False

        tmp.add(node)
        container = node.parent_component()
        if container is node:
            node = node.parent_block()
        else:
            node = container


def _to_dict(val):
    if isinstance(val, (dict, ComponentMap)):
        return val
    return {None: val}


def get_src_disjunction(xor_constraint):
    """Return the Disjunction corresponding to xor_constraint

    Parameters
    ----------
    xor_constraint: Constraint, which must be the logical constraint
                    (located on the transformation block) of some
                    Disjunction
    """
    # NOTE: This is indeed a linear search through the Disjunctions on the
    # model. I am leaving it this way on the assumption that asking XOR
    # constraints for their Disjunction is not going to be a common
    # question. If we ever need efficiency then we should store a reverse
    # map from the XOR constraint to the Disjunction on the transformation
    # block while we do the transformation. And then this method could query
    # that map.
    m = xor_constraint.model()
    for disjunction in m.component_data_objects(
        Disjunction, descend_into=(Block, Disjunct)
    ):
        if disjunction._algebraic_constraint:
            if disjunction._algebraic_constraint() is xor_constraint:
                return disjunction
    raise GDP_Error(
        "It appears that '%s' is not an XOR or OR constraint "
        "resulting from transforming a Disjunction." % xor_constraint.name
    )


def get_src_disjunct(transBlock):
    """Return the Disjunct object whose transformed components are on
    transBlock.

    Parameters
    ----------
    transBlock: BlockData which is in the relaxedDisjuncts IndexedBlock
                on a transformation block.
    """
    if (
        not hasattr(transBlock, "_src_disjunct")
        or type(transBlock._src_disjunct) is not weakref_ref
    ):
        raise GDP_Error(
            "Block '%s' doesn't appear to be a transformation "
            "block for a disjunct. No source disjunct found." % transBlock.name
        )
    return transBlock._src_disjunct()


def get_src_constraint(transformedConstraint):
    """Return the original Constraint whose transformed counterpart is
    transformedConstraint

    Parameters
    ----------
    transformedConstraint: Constraint, which must be a component on one of
    the BlockDatas in the relaxedDisjuncts Block of
    a transformation block
    """
    transBlock = transformedConstraint.parent_block()
    src_constraints = transBlock.private_data('pyomo.gdp').src_constraint
    # This should be our block, so if it's not, the user messed up and gave
    # us the wrong thing. If they happen to also have a _constraintMap then
    # the world is really against us.
    if transformedConstraint not in src_constraints:
        raise GDP_Error(
            "Constraint '%s' is not a transformed constraint"
            % transformedConstraint.name
        )
    # if something goes wrong here, it's a bug in the mappings.
    return src_constraints[transformedConstraint]


def _find_parent_disjunct(constraint):
    # traverse up until we find the disjunct this constraint lives on
    parent_disjunct = constraint.parent_block()
    while not isinstance(parent_disjunct, DisjunctData):
        if parent_disjunct is None:
            raise GDP_Error(
                "Constraint '%s' is not on a disjunct and so was not "
                "transformed" % constraint.name
            )
        parent_disjunct = parent_disjunct.parent_block()

    return parent_disjunct


def _get_constraint_transBlock(constraint):
    parent_disjunct = _find_parent_disjunct(constraint)
    # we know from _find_parent_disjunct that parent_disjunct is a Disjunct,
    # so the below is OK
    transBlock = parent_disjunct._transformation_block
    if transBlock is None:
        raise GDP_Error(
            "Constraint '%s' is on a disjunct which has not been "
            "transformed" % constraint.name
        )
    # if it's not None, it's the weakref we wanted.
    transBlock = transBlock()

    return transBlock


def get_transformed_constraints(srcConstraint):
    """Return the transformed version of srcConstraint

    Parameters
    ----------
    srcConstraint: ScalarConstraint or ConstraintData, which must be in
    the subtree of a transformed Disjunct
    """
    if srcConstraint.is_indexed():
        raise GDP_Error(
            "Argument to get_transformed_constraint should be "
            "a ScalarConstraint or ConstraintData. (If you "
            "want the container for all transformed constraints "
            "from an IndexedDisjunction, this is the parent "
            "component of a transformed constraint originating "
            "from any of its ComponentDatas.)"
        )
    transBlock = _get_constraint_transBlock(srcConstraint)
    transformed_constraints = transBlock.private_data(
        'pyomo.gdp'
    ).transformed_constraints
    if srcConstraint in transformed_constraints:
        return transformed_constraints[srcConstraint]
    else:
        raise GDP_Error(
            "Constraint '%s' has not been transformed." % srcConstraint.name
        )


def _warn_for_active_disjunct(innerdisjunct, outerdisjunct):
    assert innerdisjunct.active
    problemdisj = innerdisjunct
    if innerdisjunct.is_indexed():
        for i in sorted(innerdisjunct.keys()):
            if innerdisjunct[i].active:
                # This shouldn't be true, we will complain about it.
                problemdisj = innerdisjunct[i]
                break

    raise GDP_Error(
        "Found active disjunct '{0}' in disjunct '{1}'! Either {0} "
        "is not in a disjunction or the disjunction it is in "
        "has not been transformed. {0} needs to be deactivated "
        "or its disjunction transformed before {1} can be "
        "transformed.".format(
            problemdisj.getname(fully_qualified=True),
            outerdisjunct.getname(fully_qualified=True),
        )
    )


def check_model_algebraic(instance):
    """Checks if there are any active Disjuncts or Disjunctions reachable via
    active Blocks. If there are not, it returns True. If there are, it issues
    a warning detailing where in the model there are remaining non-algebraic
    components, and returns False.

    Parameters
    ----------
    instance: a Model or Block
    """
    disjunction_set = {
        i
        for i in instance.component_data_objects(
            Disjunction, descend_into=(Block, Disjunct), active=None
        )
    }
    active_disjunction_set = {
        i
        for i in instance.component_data_objects(
            Disjunction, descend_into=(Block, Disjunct), active=True
        )
    }
    disjuncts_in_disjunctions = set()
    for i in disjunction_set:
        disjuncts_in_disjunctions.update(i.disjuncts)
    disjuncts_in_active_disjunctions = set()
    for i in active_disjunction_set:
        disjuncts_in_active_disjunctions.update(i.disjuncts)

    for disjunct in instance.component_data_objects(
        Disjunct, descend_into=(Block,), descent_order=TraversalStrategy.PostfixDFS
    ):
        # check if it's relaxed
        if disjunct.transformation_block is not None:
            continue
        # It's not transformed, check if we should complain
        elif (
            disjunct.active
            and _disjunct_not_fixed_true(disjunct)
            and _disjunct_on_active_block(disjunct)
        ):
            # If someone thinks they've transformed the whole instance, but
            # there is still an active Disjunct on the model, we will warn
            # them. In the future this should be the writers' job.)
            if disjunct not in disjuncts_in_disjunctions:
                logger.warning(
                    'Disjunct "%s" is currently active, '
                    'but was not found in any Disjunctions. '
                    'This is generally an error as the model '
                    'has not been fully relaxed to a '
                    'pure algebraic form.' % (disjunct.name,)
                )
                return False
            elif disjunct not in disjuncts_in_active_disjunctions:
                logger.warning(
                    'Disjunct "%s" is currently active. While '
                    'it participates in a Disjunction, '
                    'that Disjunction is currently deactivated. '
                    'This is generally an error as the '
                    'model has not been fully relaxed to a pure '
                    'algebraic form. Did you deactivate '
                    'the Disjunction without addressing the '
                    'individual Disjuncts?' % (disjunct.name,)
                )
                return False
            else:
                logger.warning(
                    'Disjunct "%s" is currently active. It must be '
                    'transformed or deactivated before solving the '
                    'model.' % (disjunct.name,)
                )
                return False

    for cons in instance.component_data_objects(
        LogicalConstraint, descend_into=Block, active=True
    ):
        if cons.active:
            logger.warning(
                'LogicalConstraint "%s" is currently active. It '
                'must be transformed or deactivated before solving '
                'the model.' % cons.name
            )
            return False

    # We didn't find anything bad.
    return True


def _disjunct_not_fixed_true(disjunct):
    # Return true if the disjunct indicator variable is not fixed to True
    return not (disjunct.indicator_var.fixed and disjunct.indicator_var.value)


def _disjunct_on_active_block(disjunct):
    # Check first to make sure that the disjunct is not a descendent of an
    # inactive Block or fixed and deactivated Disjunct, before raising a
    # warning.
    parent_block = disjunct.parent_block()
    while parent_block is not None:
        # deactivated Block
        if parent_block.ctype is Block and not parent_block.active:
            return False
        # properly deactivated Disjunct
        elif (
            parent_block.ctype is Disjunct
            and not parent_block.active
            and parent_block.indicator_var.value == False
            and parent_block.indicator_var.fixed
        ):
            return False
        else:
            # Step up one level in the hierarchy
            parent_block = parent_block.parent_block()
            continue
    return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from .doe import DesignOfExperiments, ObjectiveLib, FiniteDifferenceStep
from .utils import rescale_FIM
from .grey_box_utilities import FIMExternalGreyBox

# Deprecation errors for old Pyomo.DoE interface classes and structures
from pyomo.common.deprecation import deprecated

deprecation_message = (
    "Pyomo.DoE has been refactored. The current interface utilizes Experiment "
    "objects that label unknown parameters, experiment inputs, experiment outputs "
    "and measurement error. This avoids fragile string-based naming. For "
    "instructions on using the new interface, please see the Pyomo.DoE documentation "
    "`https://pyomo.readthedocs.io/en/latest/explanation/analysis/doe/doe.html`"
)


@deprecated(
    "Use of MeasurementVariables in Pyomo.DoE is no longer supported.", version='6.8.0'
)
class MeasurementVariables:
    def __init__(self, *args):
        raise RuntimeError(deprecation_message)


@deprecated(
    "Use of DesignVariables in Pyomo.DoE is no longer supported.", version='6.8.0'
)
class DesignVariables:
    def __init__(self, *args):
        raise RuntimeError(deprecation_message)


@deprecated(
    "Use of ModelOptionLib in Pyomo.DoE is no longer supported.", version='6.8.0'
)
class ModelOptionLib:
    def __init__(self, *args):
        raise RuntimeError(deprecation_message)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import collections
from .visitor import StreamBasedExpressionVisitor
from .numvalue import nonpyomo_leaf_types
from pyomo.core.expr import (
    LinearExpression,
    MonomialTermExpression,
    SumExpression,
    ExpressionBase,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    NegationExpression,
    UnaryFunctionExpression,
    ExternalFunctionExpression,
    NPV_ProductExpression,
    NPV_DivisionExpression,
    NPV_PowExpression,
    NPV_SumExpression,
    NPV_NegationExpression,
    NPV_UnaryFunctionExpression,
    NPV_ExternalFunctionExpression,
    Expr_ifExpression,
    AbsExpression,
    NPV_AbsExpression,
    NumericValue,
    RangedExpression,
    InequalityExpression,
    EqualityExpression,
    GetItemExpression,
)
from typing import List
from pyomo.common.collections import Sequence
from pyomo.common.errors import PyomoException
from pyomo.common.formatting import tostr
from pyomo.common.numeric_types import native_types


def handle_expression(node: ExpressionBase, pn: List):
    pn.append((type(node), node.nargs()))
    return node.args


def handle_named_expression(node, pn: List, include_named_exprs=True):
    if include_named_exprs:
        pn.append((type(node), 1))
    return (node.expr,)


def handle_unary_expression(node: UnaryFunctionExpression, pn: List):
    pn.append((type(node), 1, node.getname()))
    return node.args


def handle_external_function_expression(node: ExternalFunctionExpression, pn: List):
    pn.append((type(node), node.nargs(), node._fcn))
    return node.args


def handle_sequence(node: collections.abc.Sequence, pn: List):
    pn.append((collections.abc.Sequence, len(node)))
    return list(node)


def handle_inequality(node: collections.abc.Sequence, pn: List):
    pn.append((type(node), node.nargs(), node.strict))
    return node.args


def _generic_expression_handler():
    return handle_expression


handler = collections.defaultdict(_generic_expression_handler)

handler[UnaryFunctionExpression] = handle_unary_expression
handler[NPV_UnaryFunctionExpression] = handle_unary_expression
handler[ExternalFunctionExpression] = handle_external_function_expression
handler[NPV_ExternalFunctionExpression] = handle_external_function_expression
handler[AbsExpression] = handle_unary_expression
handler[NPV_AbsExpression] = handle_unary_expression
handler[InequalityExpression] = handle_inequality
handler[RangedExpression] = handle_inequality
handler[list] = handle_sequence


class PrefixVisitor(StreamBasedExpressionVisitor):
    def __init__(self, include_named_exprs=True):
        super().__init__()
        self._result = None
        self._include_named_exprs = include_named_exprs

    def initializeWalker(self, expr):
        self._result = []
        return True, None

    def enterNode(self, node):
        ntype = type(node)
        if ntype in nonpyomo_leaf_types:
            self._result.append(node)
            return tuple(), None

        if ntype in handler:
            return handler[ntype](node, self._result), None

        if hasattr(node, 'is_expression_type'):
            if node.is_expression_type():
                if node.is_named_expression_type():
                    return (
                        handle_named_expression(
                            node, self._result, self._include_named_exprs
                        ),
                        None,
                    )
                else:
                    return handler[ntype](node, self._result), None
        elif hasattr(node, '__len__'):
            handler[ntype] = handle_sequence
            return handle_sequence(node, self._result), None

        self._result.append(node)
        return tuple(), None

    def finalizeResult(self, result):
        ans = self._result
        self._result = None
        return ans


def convert_expression_to_prefix_notation(expr, include_named_exprs=True):
    """
    This function converts pyomo expressions to a list that looks very
    much like prefix notation.  The result can be used in equality
    comparisons to compare expression trees.

    Note that the data structure returned by this function might be
    changed in the future. However, we will maintain that the result
    can be used in equality comparisons.

    Also note that the result should really only be used in equality
    comparisons if the equality comparison is expected to return
    True. If the expressions being compared are expected to be
    different, then the equality comparison will often result in an
    error rather than returning False.

    m = ConcreteModel()
    m.x = Var()
    m.y = Var()

    e1 = m.x * m.y
    e2 = m.x * m.y
    e3 = m.x + m.y

    convert_expression_to_prefix_notation(e1) == convert_expression_to_prefix_notation(e2)  # True
    convert_expression_to_prefix_notation(e1) == convert_expression_to_prefix_notation(e3)  # Error

    However, the compare_expressions function can be used:

    compare_expressions(e1, e2)  # True
    compare_expressions(e1, e3)  # False

    Parameters
    ----------
    expr: NumericValue
        A Pyomo expression, Var, or Param

    Returns
    -------
    prefix_notation: list
        The expression in prefix notation

    """
    visitor = PrefixVisitor(include_named_exprs=include_named_exprs)
    return visitor.walk_expression(expr)


def compare_expressions(expr1, expr2, include_named_exprs=True):
    """Returns True if 2 expression trees are identical, False otherwise.

    Parameters
    ----------
    expr1: NumericValue
        A Pyomo Var, Param, or expression
    expr2: NumericValue
        A Pyomo Var, Param, or expression
    include_named_exprs: bool
        If False, then named expressions will be ignored. In other
        words, this function will return True if one expression has a
        named expression and the other does not as long as the rest of
        the expression trees are identical.

    Returns
    -------
    res: bool
        A bool indicating whether or not the expressions are identical.

    """
    pn1 = convert_expression_to_prefix_notation(
        expr1, include_named_exprs=include_named_exprs
    )
    pn2 = convert_expression_to_prefix_notation(
        expr2, include_named_exprs=include_named_exprs
    )
    try:
        res = pn1 == pn2
    except (PyomoException, AttributeError):
        res = False
    return res


def assertExpressionsEqual(test, a, b, include_named_exprs=True, places=None):
    """unittest-based assertion for comparing expressions

    This converts the expressions `a` and `b` into prefix notation and
    then compares the resulting lists.

    Parameters
    ----------
    test: unittest.TestCase
        The unittest `TestCase` class that is performing the test.

    a: ExpressionBase or native type

    b: ExpressionBase or native type

    include_named_exprs : bool
        If True (the default), the comparison expands all named
        expressions when generating the prefix notation

    places : int
        Number of decimal places required for equality of floating
        point numbers in the expression. If None (the default), the
        expressions must be exactly equal.
    """
    prefix_a = convert_expression_to_prefix_notation(a, include_named_exprs)
    prefix_b = convert_expression_to_prefix_notation(b, include_named_exprs)
    try:
        test.assertEqual(len(prefix_a), len(prefix_b))
        for _a, _b in zip(prefix_a, prefix_b):
            test.assertIs(_a.__class__, _b.__class__)
            # If _a is nan, check _b is nan
            if _a != _a:
                test.assertTrue(_b != _b)
            else:
                if places is None:
                    test.assertEqual(_a, _b)
                else:
                    test.assertAlmostEqual(_a, _b, places=places)
    except (PyomoException, AssertionError):
        test.fail(
            f"Expressions not equal:\n\t"
            f"{tostr(prefix_a)}\n\t!=\n\t{tostr(prefix_b)}"
        )


def assertExpressionsStructurallyEqual(
    test, a, b, include_named_exprs=True, places=None
):
    """unittest-based assertion for comparing expressions

    This converts the expressions `a` and `b` into prefix notation and
    then compares the resulting lists.  Operators and (non-native type)
    leaf nodes in the prefix representation are converted to strings
    before comparing (so that things like variables can be compared
    across clones or pickles)

    Parameters
    ----------
    test: unittest.TestCase
        The unittest `TestCase` class that is performing the test.

    a: ExpressionBase or native type

    b: ExpressionBase or native type

    include_named_exprs : bool
       If True (the default), the comparison expands all named
       expressions when generating the prefix notation

    places : int
        Number of decimal places required for equality of floating
        point numbers in the expression. If None (the default), the
        expressions must be exactly equal.
    """
    prefix_a = convert_expression_to_prefix_notation(a, include_named_exprs)
    prefix_b = convert_expression_to_prefix_notation(b, include_named_exprs)
    # Convert leaf nodes and operators to their string equivalents
    for prefix in (prefix_a, prefix_b):
        for i, v in enumerate(prefix):
            if type(v) in native_types:
                continue
            if type(v) is tuple:
                # This is an expression node.  Most expression nodes are
                # 2-tuples (node type, nargs), but some are 3-tuples
                # with supplemental data.  The biggest problem is
                # external functions, where the third element is the
                # external function.  We need to convert that to a
                # string to support "structural" comparisons.
                if len(v) == 3:
                    prefix[i] = v[:2] + (str(v[2]),)
                continue
            # This should be a leaf node (Var, mutable Param, etc.).
            # Convert to string to support "structural" comparison
            # (e.g., across clones)
            prefix[i] = str(v)
    try:
        test.assertEqual(len(prefix_a), len(prefix_b))
        for _a, _b in zip(prefix_a, prefix_b):
            if _a.__class__ not in native_types and _b.__class__ not in native_types:
                test.assertIs(_a.__class__, _b.__class__)
            if _a != _a:
                test.assertTrue(_b != _b)
            else:
                if places is None:
                    test.assertEqual(_a, _b)
                else:
                    test.assertAlmostEqual(_a, _b, places=places)
    except (PyomoException, AssertionError):
        test.fail(
            f"Expressions not structurally equal:\n\t"
            f"{tostr(prefix_a)}\n\t!=\n\t{tostr(prefix_b)}"
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Example in the online doc.

The expected optimal solution value is 2.438447187191098.

    Problem type:    convex MINLP
            size:    1  binary variable
                     1  continuous variables
                     2  constraints

"""

from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    Objective,
    Var,
    minimize,
    log,
)
from pyomo.common.collections import ComponentMap


class OnlineDocExample(ConcreteModel):
    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'OnlineDocExample')
        super(OnlineDocExample, self).__init__(*args, **kwargs)
        m = self
        m.x = Var(bounds=(1.0, 10.0), initialize=5.0)
        m.y = Var(within=Binary)
        m.c1 = Constraint(expr=(m.x - 4.0) ** 2 - m.x <= 50.0 * (1 - m.y))
        m.c2 = Constraint(expr=m.x * log(m.x) + 5 <= 50.0 * (m.y))
        m.objective = Objective(expr=m.x, sense=minimize)
        m.optimal_value = 2.438447
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.x] = 2.4384471855377243
        m.optimal_solution[m.y] = 1.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np, pandas as pd
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    ReactorDesignExperiment,
)


def main():

    # Read in data
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, "reactor_data.csv"))
    data = pd.read_csv(file_name)

    seed = 524  # Set seed for reproducibility
    np.random.seed(seed)  # Set seed for reproducibility

    # Create more data for the example
    N = 50
    df_std = data.std().to_frame().transpose()
    df_rand = pd.DataFrame(np.random.normal(size=N))
    df_sample = data.sample(N, replace=True).reset_index(drop=True)
    data = df_sample + df_rand.dot(df_std) / 10

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(ReactorDesignExperiment(data, i))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    pest = parmest.Estimator(exp_list, obj_function='SSE')

    # Parameter estimation
    obj, theta = pest.theta_est()
    print(obj)
    print(theta)

    ### Parameter estimation with 'leave-N-out'
    # Example use case: For each combination of data where one data point is left
    # out, estimate theta
    lNo_theta = pest.theta_est_leaveNout(1, seed=524)
    print(lNo_theta.head())

    parmest.graphics.pairwise_plot(lNo_theta, theta, seed=524)

    ### Leave one out/boostrap analysis
    # Example use case: leave 25 data points out, run 20 bootstrap samples with the
    # remaining points, determine if the theta estimate using the points left out
    # is inside or outside an alpha region based on the bootstrap samples, repeat
    # 5 times. Results are stored as a list of tuples, see API docs for information.
    lNo = 25
    lNo_samples = 5
    bootstrap_samples = 20
    dist = "MVN"
    alphas = [0.7, 0.8, 0.9]

    results = pest.leaveNout_bootstrap_test(
        lNo, lNo_samples, bootstrap_samples, dist, alphas, seed=524
    )

    # Plot results for a single value of alpha
    alpha = 0.8
    for i in range(lNo_samples):
        theta_est_N = results[i][1]
        bootstrap_results = results[i][2]
        parmest.graphics.pairwise_plot(
            bootstrap_results,
            theta_est_N,
            alpha,
            ["MVN"],
            title="Alpha: " + str(alpha) + ", " + str(theta_est_N.loc[0, alpha]),
            seed=524 + i,  # setting the seed for testing repeatability,
            # for typical use cases, this should not be set
        )

    # Extract the percent of points that are within the alpha region
    r = [results[i][1].loc[0, alpha] for i in range(lNo_samples)]
    percent_true = sum(r) / len(r)
    print(percent_true)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the BARON interface."""

from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.environ import ConcreteModel, Constraint, Objective, Var, log10, minimize
from pyomo.opt import SolverFactory, TerminationCondition

# check if BARON is available
from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

baron_available = _test_solver_cases('baron', 'bar').available


@unittest.skipIf(not baron_available, "The 'BARON' solver is not available")
@unittest.pytest.mark.solver("baron")
class BaronTest(unittest.TestCase):
    """Test the BARON interface."""

    def test_log10(self):
        # Tests the special transformation for log10
        with SolverFactory("baron") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=log10(m.x) >= 2)
            m.obj = Objective(expr=m.x, sense=minimize)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    def test_abs(self):
        # Tests the special transformation for abs
        with SolverFactory("baron") as opt:
            m = ConcreteModel()
            m.x = Var(bounds=(-100, 1))
            m.c = Constraint(expr=abs(m.x) >= 2)
            m.obj = Objective(expr=m.x, sense=minimize)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    def test_pow(self):
        # Tests the special transformation for x ^ y (both variables)
        with SolverFactory("baron") as opt:
            m = ConcreteModel()
            m.x = Var(bounds=(10, 100))
            m.y = Var(bounds=(1, 10))
            m.c = Constraint(expr=m.x**m.y >= 20)
            m.obj = Objective(expr=m.x, sense=minimize)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    def test_BARON_option_warnings(self):
        os = StringIO()
        with LoggingIntercept(os, 'pyomo.solvers'):
            m = ConcreteModel()
            m.x = Var()
            m.obj = Objective(expr=m.x**2)

            with SolverFactory("baron") as opt:
                results = opt.solve(
                    m, options={'ResName': 'results.lst', 'TimName': 'results.tim'}
                )

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )
        self.assertIn(
            'Ignoring user-specified option "ResName=results.lst"', os.getvalue()
        )
        self.assertIn(
            'Ignoring user-specified option "TimName=results.tim"', os.getvalue()
        )


if __name__ == '__main__':
    unittest.main()
